#!/usr/bin/env python3
#
# mmgen = Multi-Mode GENerator, command-line Bitcoin cold storage solution
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# ------------------------------------------------------------------------------
# MMGen note: this is a sorted version of the wordlist. The original can be
# found in the file 'wordlists/mn_wordlist.c' of the mmgen distribution.
# ------------------------------------------------------------------------------
# Oren Tirosh is no longer maintaining the original version of this
# project.
#
# Stephen Paul Weber likes it and is making it accessible on GitHub.
# https://github.com/singpolyma/mnemonicode
# ------------------------------------------------------------------------------
# Original copyright notice follows:
# ------------------------------------------------------------------------------
# Oren Tirosh wordlist
# Copyright (c) 2000  Oren Tirosh <oren@hishome.net>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

words = tuple("""
abraham
absent
absorb
academy
accent
acid
acrobat
action
active
actor
adam
address
adios
admiral
adrian
africa
agatha
agenda
agent
airline
airport
alabama
aladdin
alamo
alarm
alaska
albert
albino
album
alcohol
alert
alex
alfonso
alfred
algebra
alias
alibi
alice
alien
almanac
almond
aloha
alpha
alpine
amadeus
amanda
amazon
amber
ambient
amen
america
amigo
ammonia
analog
analyze
anatomy
andrea
andy
angel
animal
anita
annual
answer
antenna
antonio
anvil
apollo
appear
apple
april
apropos
arcade
archer
archive
arctic
arena
ariel
arizona
armada
armor
arnold
aroma
arrow
arsenal
arthur
artist
asia
aspect
aspirin
athena
athlete
atlanta
atlas
atomic
audio
august
aurora
austin
austria
avalon
avatar
avenue
average
axiom
axis
aztec
baboon
baby
bagel
baggage
bahama
baker
balance
bali
ballad
ballet
balloon
balsa
bambino
bamboo
banana
bandit
banjo
bank
barbara
barcode
baron
basic
basil
basket
battery
bazaar
bazooka
beach
beast
beatles
beauty
before
begin
belgium
benefit
benny
berlin
bermuda
bernard
betty
between
beyond
bicycle
bikini
billy
binary
bingo
biology
biscuit
bishop
bison
blast
bless
blitz
block
blonde
blue
bogart
bombay
bonanza
bonjour
bonus
book
border
boris
boston
bottle
boxer
brain
brandy
brave
bravo
brazil
bread
break
brenda
bridge
brigade
british
broken
bronze
brother
brown
bruce
bruno
brush
bucket
budget
buenos
buffalo
bundle
burger
burma
button
buzzer
byte
cabaret
cabinet
cable
cactus
cadet
caesar
cafe
cairo
cake
calypso
camel
camera
camilla
campus
canada
canal
canary
candid
candle
cannon
canoe
cantina
canvas
canyon
capital
capitan
capsule
caramel
caravan
carbon
career
cargo
carlo
carmen
carol
carpet
carrot
cartel
cartoon
casino
castle
castro
catalog
cave
caviar
cecilia
cello
celtic
cement
center
century
ceramic
chamber
chance
change
channel
chant
chaos
chapter
chariot
charlie
charm
charter
cheese
chef
chemist
cherry
chess
chicago
chicken
chief
child
china
choice
chris
chrome
cigar
cinema
cipher
circle
circus
citizen
citrus
city
civil
clara
clarion
clark
classic
claudia
clean
clever
client
cliff
climax
clinic
clock
clone
cloud
club
cobalt
cobra
cockpit
coconut
cola
collect
college
colombo
colony
color
combat
comedy
comet
command
common
compact
company
compare
compass
complex
comrade
conan
concept
concert
condor
conduct
congo
connect
consul
contact
content
context
contour
control
convert
cool
copper
copy
coral
corner
corona
correct
cosmos
costume
cotton
couple
courage
cover
cowboy
crack
craft
crash
crater
credit
cricket
crimson
critic
crown
crystal
cuba
cubic
culture
cupid
current
cycle
cyclone
dallas
dance
daniel
danube
darwin
data
david
deal
decade
decide
decimal
declare
degree
delete
deliver
delphi
delta
deluxe
demand
demo
denmark
denver
depend
derby
desert
design
desire
detail
detect
develop
dexter
diagram
dialog
diamond
diana
diego
diesel
diet
digital
dilemma
dinner
diploma
direct
disco
disney
dispute
distant
divide
doctor
dolby
dollar
domain
domingo
dominic
domino
donald
donor
door
double
dragon
drama
dream
drink
driver
druid
drum
dublin
duet
dynamic
dynasty
eagle
earth
east
easy
echo
eclipse
ecology
economy
eddie
edgar
edison
edition
editor
educate
edward
effect
ego
egypt
elastic
electra
elegant
element
elite
elvis
email
emerald
emotion
empire
empty
energy
engine
english
enigma
enjoy
enrico
episode
epoxy
equal
equator
eric
erosion
escape
escort
eternal
ethnic
europe
evening
event
everest
evident
evita
exact
example
except
exhibit
exile
exit
exodus
exotic
expand
explain
explore
export
express
extend
extra
extreme
fabric
factor
falcon
fame
family
famous
fantasy
farmer
fashion
fast
father
fax
felix
ferrari
fiber
fiction
fidel
field
fiesta
figure
film
filter
final
finance
finish
finland
fiona
fire
first
fish
flag
flame
flash
flex
flipper
float
flood
floor
florida
flower
fluid
flute
focus
folio
food
forbid
ford
forest
forever
forget
formal
format
formula
fortune
forum
forward
fossil
fractal
fragile
frame
france
frank
freddie
freedom
fresh
friday
friend
frog
front
frozen
fruit
fuel
fuji
future
gabriel
galaxy
galileo
gallery
gallop
game
gamma
garage
garbo
garcia
garden
garlic
gate
gemini
general
genesis
genetic
geneva
genius
gentle
george
germany
giant
gibson
gilbert
ginger
giraffe
gizmo
glass
global
gloria
goblin
gold
golf
gondola
gong
good
gopher
gordon
gorilla
gossip
grace
gram
grand
granite
graph
gravity
gray
greek
green
gregory
grid
griffin
grille
ground
group
guest
guide
guitar
guru
gustav
gyro
habitat
hair
halt
hamlet
hammer
hand
happy
harbor
harlem
harmony
harris
harvard
harvest
havana
hawaii
hazard
heart
heaven
heavy
helena
helium
hello
henry
herbert
herman
heroic
hexagon
hilton
hippie
history
hobby
holiday
honey
hope
horizon
horse
hostel
hotel
house
human
humor
hunter
husband
hydro
ibiza
iceberg
icon
idea
igloo
igor
image
imagine
imitate
immune
impact
import
inca
inch
index
india
indigo
infant
info
ingrid
initial
input
insect
inside
instant
invent
invest
invite
iris
iron
isabel
isotope
italian
ivan
ivory
jacket
jackson
jacob
jaguar
jamaica
james
janet
japan
jargon
jasmine
jason
java
jazz
jeep
jerome
jessica
jester
jet
jimmy
job
joel
john
join
joker
jordan
joseph
joshua
journal
judge
judo
juice
juliet
julius
july
jumbo
jump
june
jungle
junior
jupiter
justice
justin
kansas
karate
karl
karma
kayak
kermit
kevin
kilo
kimono
kinetic
king
kitchen
kiwi
koala
korea
labor
ladder
lady
lagoon
lake
laptop
laser
latin
laura
lava
lazarus
learn
lecture
left
legacy
legal
legend
lemon
leonid
lesson
letter
level
lexicon
liberal
libra
license
life
light
lima
limbo
limit
linda
linear
lion
liquid
list
liter
lithium
little
llama
lobby
lobster
local
locate
logic
logo
lola
london
lopez
lorenzo
lotus
love
loyal
lucas
lucky
lunar
lunch
machine
macro
madam
madonna
madrid
maestro
magenta
magic
magnet
magnum
mailbox
major
malta
mama
mambo
mammal
manager
mango
manila
manual
marble
marco
margo
marina
marion
market
mars
martin
marvin
mary
mask
master
match
matrix
maximum
maxwell
mayday
mayor
maze
meaning
media
medical
medusa
mega
melody
melon
member
memo
memphis
mental
mentor
menu
mercury
mercy
message
metal
meteor
meter
method
metro
mexico
miami
michael
micro
middle
miguel
mike
milan
mile
milk
miller
million
mimic
mimosa
mineral
minimum
minus
minute
miracle
mirage
miranda
mirror
mission
mister
mixer
mobile
model
modem
modern
modest
modular
moment
monaco
monarch
monday
money
monica
monitor
monkey
mono
monster
montana
moral
morgan
morning
morph
morris
moses
motel
mother
motif
motor
mouse
mozart
multi
museum
music
mustang
mystery
nadia
nancy
natasha
native
nato
natural
navy
nebula
nectar
needle
nelson
neon
nepal
neptune
nerve
network
neuron
neutral
nevada
never
news
newton
next
nice
nickel
night
nikita
nina
ninja
nirvana
nissan
nitro
nixon
nobel
nobody
noise
nominal
normal
north
norway
nothing
nova
novel
nuclear
null
number
numeric
nurse
nylon
oasis
oberon
object
observe
ocean
octavia
october
octopus
office
ohio
olga
oliver
olivia
olympic
omega
open
opera
opinion
optic
optimal
option
opus
orange
orbit
orca
orchid
order
oregano
organic
orient
origami
origin
orinoco
orion
orlando
oscar
othello
outside
oval
owner
oxford
oxygen
ozone
pablo
pacific
package
page
pagoda
paint
palace
palma
pamela
panama
pancake
panda
pandora
panel
panic
panther
papa
paper
paprika
parade
paradox
pardon
parent
paris
parker
parking
parody
parole
partner
passage
passive
pasta
pastel
patent
patient
patriot
patrol
patron
pattern
paul
peace
pearl
pedro
pegasus
pelican
pencil
penguin
people
pepper
percent
perfect
perform
perfume
period
permit
person
peru
phantom
philips
phoenix
phone
photo
phrase
piano
picasso
picnic
picture
pierre
pigment
pilgrim
pilot
pinball
pioneer
pirate
pixel
pizza
place
planet
plasma
plaster
plastic
plate
plato
plaza
plume
pluto
pocket
podium
poem
poetic
pogo
point
poker
polaris
police
polite
politic
polka
polo
polygon
poncho
pony
popcorn
popular
portal
postage
postal
potato
powder
prague
precise
prefix
prelude
premium
prepare
present
press
presto
pretend
pretty
price
prime
prince
printer
prism
private
prize
process
product
profile
profit
program
project
promise
promo
protect
protein
proton
provide
proxy
public
pulse
puma
pump
pupil
puzzle
pyramid
python
quality
quarter
quasi
quebec
queen
quest
quick
quiet
quiz
quota
rabbit
race
rachel
radar
radical
radio
radius
rainbow
raja
ralph
ramirez
random
ranger
rapid
ravioli
raymond
rebel
record
recycle
reflex
reform
regard
region
regular
relax
remark
remote
rent
repair
reply
report
reptile
reserve
respect
respond
result
resume
retro
reunion
reverse
reward
rhino
ribbon
ricardo
richard
rider
right
ringo
rio
risk
ritual
rival
river
riviera
road
robert
robin
robot
rocket
rodent
rodeo
roger
roman
romeo
rondo
roof
rose
round
rover
royal
rubber
ruby
rudolf
rufus
russian
sabine
sabrina
saddle
safari
saga
sahara
sailor
saint
salad
salami
salary
salmon
salon
salsa
salt
salute
samba
sample
samuel
sandra
santana
sardine
satire
saturn
savage
scale
scarlet
scholar
school
scoop
scorpio
scratch
screen
script
scroll
scuba
season
second
secret
section
sector
secure
segment
select
seminar
senator
senior
sensor
serial
serpent
service
shadow
shake
shallow
shampoo
shannon
sharon
sharp
shave
shelf
shelter
sheriff
sherman
shine
ship
shirt
shock
shoe
short
shrink
side
sierra
sigma
signal
silence
silicon
silk
silver
similar
simon
simple
sinatra
sincere
singer
single
siren
sister
size
ski
slalom
slang
sleep
slogan
slow
small
smart
smile
smoke
snake
snow
social
society
soda
sofia
solar
solid
solo
sonar
sonata
song
sonic
soprano
sound
source
south
soviet
spain
spark
sparta
special
speech
speed
spell
spend
sphere
spider
spiral
spirit
split
sponsor
spoon
sport
spray
spring
square
stadium
stage
stamp
stand
star
state
static
station
status
stella
stereo
stick
sting
stock
stone
stop
store
storm
story
strange
street
stretch
strong
stuart
student
studio
style
subject
subway
sugar
sulfur
sultan
summer
sunday
sunset
super
support
survive
susan
sushi
suzuki
sweden
sweet
swim
swing
switch
symbol
system
table
taboo
tactic
tahiti
talent
tango
tape
target
tarzan
taxi
teacher
telecom
telex
temple
tempo
tennis
texas
textile
theory
thermos
think
thomas
tibet
ticket
tictac
tiger
time
tina
titanic
toast
tobacco
today
toga
tokyo
tomato
tommy
tonight
topic
torch
tornado
toronto
torpedo
torso
total
totem
touch
tourist
tower
toyota
tractor
trade
traffic
transit
trapeze
travel
tribal
tribune
trick
trident
trilogy
trinity
tripod
triton
trivial
tropic
truck
trumpet
trust
tulip
tuna
tunnel
turbo
turtle
twin
twist
type
ultra
uncle
under
unicorn
uniform
union
unique
unit
update
uranium
urban
urgent
user
vacuum
valery
valid
value
vampire
vanilla
vatican
vega
velvet
vendor
venice
ventura
venus
verbal
verona
version
vertigo
veteran
vibrate
victor
video
vienna
viking
village
vincent
violet
violin
virgo
virtual
virus
visa
visible
vision
visitor
vista
visual
vital
vitamin
viva
vocal
vodka
voice
volcano
voltage
volume
voodoo
vortex
voyage
waiter
warning
watch
water
wave
weather
wedding
weekend
welcome
western
wheel
whiskey
william
window
winter
wisdom
wizard
wolf
wonder
world
xray
yankee
year
yellow
yes
yoga
yogurt
young
yoyo
zebra
zero
zigzag
zipper
zodiac
zoom
""".split())
