#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>

"""
test.objtest_d.common: shared data for MMGen data objects tests
"""

from ..include.common import getrand

r16, r32 = getrand(16), getrand(32)
